import zengl


def test_frame_time(ctx: zengl.Context):
    image = ctx.image((64, 64), 'rgba8unorm')
    pipeline = ctx.pipeline(
        vertex_shader='''
            #version 330 core

            vec2 positions[3] = vec2[](
                vec2(0.1, 0.0),
                vec2(-0.05, 0.086),
                vec2(-0.05, -0.086)
            );

            void main() {
                gl_Position = vec4(positions[gl_VertexID] + 0.5, 0.0, 1.0);
            }
        ''',
        fragment_shader='''
            #version 330 core

            layout (location = 0) out vec4 out_color;

            void main() {
                out_color = vec4(0.0, 0.0, 1.0, 1.0);
            }
        ''',
        framebuffer=[image],
        topology='triangles',
        vertex_count=3,
    )

    ctx.new_frame()
    image.clear()
    pipeline.render()
    ctx.end_frame()

    assert ctx.frame_time == 0.0

    ctx.new_frame(frame_time=True)
    image.clear()
    pipeline.render()
    ctx.end_frame(sync=True)

    assert ctx.frame_time > 0.0

    ctx.new_frame()
    image.clear()
    pipeline.render()
    ctx.end_frame()

    assert ctx.frame_time == 0.0



import zengl
import numpy as np


def test_image_data(ctx: zengl.Context):
    img = ctx.image((4, 4), 'rgba8unorm', b'RGBA' * 16)
    assert img.read() == b'RGBA' * 16


def test_image_write_simple(ctx: zengl.Context):
    img = ctx.image((4, 4), 'rgba8unorm')
    img.write(b'RGBA' * 16)
    assert img.read() == b'RGBA' * 16


def test_image_write(ctx: zengl.Context):
    img = ctx.image((4, 4), 'rgba8unorm', b'RGBA' * 16)
    img.write(b'00001111', (2, 1), offset=(2, 3))
    assert img.read() == b'RGBA' * 14 + b'00001111'
    assert img.read(offset=(1, 3), size=(1, 1)) == b'RGBA'
    assert img.read(offset=(2, 3), size=(1, 1)) == b'0000'
    assert img.read(offset=(3, 3), size=(1, 1)) == b'1111'


def test_image_read_into(ctx: zengl.Context):
    img = ctx.image((4, 4), 'rgba8unorm', b'RGBA' * 16)
    img.write(b'00001111', (2, 1), offset=(2, 3))

    mem = memoryview(bytearray(64))
    img.read(into=mem)
    assert bytes(mem) == b'RGBA' * 14 + b'00001111'

    mem = memoryview(bytearray(4))
    img.read(offset=(1, 3), size=(1, 1), into=mem)
    assert bytes(mem) == b'RGBA'

    mem = memoryview(bytearray(4))
    img.read(offset=(2, 3), size=(1, 1), into=mem)
    assert bytes(mem) == b'0000'

    mem = memoryview(bytearray(4))
    img.read(offset=(3, 3), size=(1, 1), into=mem)
    assert bytes(mem) == b'1111'


def test_clear_float(ctx: zengl.Context):
    img = ctx.image((4, 4), 'rgba32float', bytearray(256))
    img.clear_value = (0.5, 0.3, 0.7, 100.0)
    img.clear()
    pixel = np.frombuffer(img.read((1, 1)), 'f4')
    np.testing.assert_array_almost_equal(pixel, [0.5, 0.3, 0.7, 100.0], decimal=3)


def test_clear_int(ctx: zengl.Context):
    img = ctx.image((4, 4), 'rgba32sint', bytearray(256))
    img.clear_value = (0, 100, -1024, 0x11223344)
    img.clear()
    pixel = np.frombuffer(img.read((1, 1)), 'i4')
    np.testing.assert_array_almost_equal(pixel, [0, 100, -1024, 0x11223344], decimal=3)


def test_clear_uint(ctx: zengl.Context):
    img = ctx.image((4, 4), 'rgba32uint', bytearray(256))
    img.clear_value = (0, 100, 1024, 0xFF000000)
    img.clear()
    pixel = np.frombuffer(img.read((1, 1)), 'u4')
    np.testing.assert_array_almost_equal(pixel, [0, 100, 1024, 0xFF000000], decimal=3)



import numpy as np
import pytest
import zengl


def test_float_image(ctx: zengl.Context):
    img = ctx.image((4, 4), 'r32float')
    img.clear_value = 0.33
    np.testing.assert_array_almost_equal([img.clear_value], [0.33], 2)
    img.clear_value = 0.75
    np.testing.assert_array_almost_equal([img.clear_value], [0.75], 2)


def test_int_image(ctx: zengl.Context):
    img = ctx.image((4, 4), 'r32sint')
    img.clear_value = 42
    np.testing.assert_array_almost_equal([img.clear_value], [42], 2)
    img.clear_value = 123456789
    np.testing.assert_array_almost_equal([img.clear_value], [123456789], 2)


def test_float_vec_image(ctx: zengl.Context):
    img = ctx.image((4, 4), 'rg32float')
    img.clear_value = 0.33, 0.66
    np.testing.assert_array_almost_equal(img.clear_value, [0.33, 0.66], 2)
    img.clear_value = 0.75, 0.12
    np.testing.assert_array_almost_equal(img.clear_value, [0.75, 0.12], 2)


def test_int_vec_image(ctx: zengl.Context):
    img = ctx.image((4, 4), 'rgba16sint')
    img.clear_value = 10, 20, 30, 40
    np.testing.assert_array_almost_equal(img.clear_value, [10, 20, 30, 40], 2)
    img.clear_value = 50, 60, 70, 80
    np.testing.assert_array_almost_equal(img.clear_value, [50, 60, 70, 80], 2)


def test_depth_stencil_image(ctx: zengl.Context):
    img = ctx.image((4, 4), 'depth24plus-stencil8')
    img.clear_value = 0.3, 200
    np.testing.assert_array_almost_equal(img.clear_value, [0.3, 200], 2)
    img.clear_value = 0.88, 12
    np.testing.assert_array_almost_equal(img.clear_value, [0.88, 12], 2)


def test_invalid_clear_value(ctx: zengl.Context):
    img_float = ctx.image((4, 4), 'r32float')
    img_int = ctx.image((4, 4), 'r32sint')
    img_depth = ctx.image((4, 4), 'depth24plus')
    img_depth_stencil = ctx.image((4, 4), 'depth24plus-stencil8')

    with pytest.raises(TypeError):
        img_float.clear_value = 100

    with pytest.raises(TypeError):
        img_float.clear_value = (100, 100)

    with pytest.raises(TypeError):
        img_float.clear_value = '100'

    with pytest.raises(TypeError):
        img_int.clear_value = 0.5

    with pytest.raises(TypeError):
        img_int.clear_value = (0.5, 0.5)

    with pytest.raises(TypeError):
        img_int.clear_value = b'0.5'

    with pytest.raises(TypeError):
        img_depth.clear_value = 2

    # with pytest.raises(TypeError):
    #     img_depth_stencil.clear_value = 2, 2

    with pytest.raises(TypeError):
        img_depth_stencil.clear_value = 0.5, 0.5



import zengl

import utils

IMAGE_FORMATS = [
    'r8unorm',
    'rg8unorm',
    'rgba8unorm',
    'r8snorm',
    'rg8snorm',
    'rgba8snorm',
    'r8uint',
    'rg8uint',
    'rgba8uint',
    'r16uint',
    'rg16uint',
    'rgba16uint',
    'r32uint',
    'rg32uint',
    'rgba32uint',
    'r8sint',
    'rg8sint',
    'rgba8sint',
    'r16sint',
    'rg16sint',
    'rgba16sint',
    'r32sint',
    'rg32sint',
    'rgba32sint',
    'r16float',
    'rg16float',
    'rgba16float',
    'r32float',
    'rg32float',
    'rgba32float',
    'depth16unorm',
    'depth24plus',
    'depth24plus-stencil8',
    'depth32float',
]


def test_texture_image_formats(ctx: zengl.Context):
    utils.clear_gl_error()
    for fmt in IMAGE_FORMATS:
        img = ctx.image((4, 4), fmt, texture=True)
        utils.assert_gl_error(fmt)
        ctx.release(img)
        utils.assert_gl_error()


def test_renderbuffer_image_formats(ctx: zengl.Context):
    utils.clear_gl_error()
    for fmt in IMAGE_FORMATS:
        img = ctx.image((4, 4), fmt, texture=False)
        utils.assert_gl_error(fmt)
        ctx.release(img)
        utils.assert_gl_error()



import pytest
import zengl


def test_vertex_shader_error(ctx: zengl.Context):
    image = ctx.image((64, 64), 'rgba8unorm')

    with pytest.raises(ValueError, match='Vertex Shader Error'):
        ctx.pipeline(
            vertex_shader='''
                #version 330 core

                void main() {
                    gl_Position = vec4(positions[gl_VertexID], 0.0, 1.0);
                }
            ''',
            fragment_shader='''
                #version 330 core

                layout (location = 0) out vec4 out_color;

                void main() {
                    out_color = vec4(0.0, 0.0, 0.0, 1.0);
                }
            ''',
            framebuffer=[image],
            topology='triangles',
            vertex_count=3,
        )


def test_fragment_shader_error(ctx: zengl.Context):
    image = ctx.image((64, 64), 'rgba8unorm')

    with pytest.raises(ValueError, match='Fragment Shader Error'):
        ctx.pipeline(
            vertex_shader='''
                #version 330 core

                vec2 positions[3] = vec2[](
                    vec2(0.1, 0.0),
                    vec2(-0.05, 0.086),
                    vec2(-0.05, -0.086)
                );

                void main() {
                    gl_Position = vec4(positions[gl_VertexID], 0.0, 1.0);
                }
            ''',
            fragment_shader='''
                #version 330 core

                layout (location = 0) out vec4 out_color;

                void main() {
                    out_color = vec3(0.0, 0.0, 0.0);
                }
            ''',
            framebuffer=[image],
            topology='triangles',
            vertex_count=3,
        )


def test_linker_error(ctx: zengl.Context):
    image = ctx.image((64, 64), 'rgba8unorm')

    with pytest.raises(ValueError, match='Linker Error'):
        ctx.pipeline(
            vertex_shader='''
                #version 330 core

                vec2 positions[3] = vec2[](
                    vec2(0.1, 0.0),
                    vec2(-0.05, 0.086),
                    vec2(-0.05, -0.086)
                );

                out vec3 v_color;

                void main() {
                    v_color = vec3(0.0, 0.0, 1.0);
                    gl_Position = vec4(positions[gl_VertexID], 0.0, 1.0);
                }
            ''',
            fragment_shader='''
                #version 330 core

                in vec4 v_color;

                layout (location = 0) out vec4 out_color;

                void main() {
                    out_color = v_color;
                }
            ''',
            framebuffer=[image],
            topology='triangles',
            vertex_count=3,
        )



import zengl


def test_image_from_buffer(ctx: zengl.Context):
    buf = ctx.buffer(b'AAAA' * 16)
    img = ctx.image((4, 4), 'rgba8unorm', buf)
    assert img.read() == b'AAAA' * 16


def test_image_write_from_buffer(ctx: zengl.Context):
    img = ctx.image((4, 4), 'rgba8unorm', b'AAAA' * 16)
    buf = ctx.buffer(b'BBBB' * 16)
    img.write(buf)
    assert img.read() == b'BBBB' * 16


def test_image_write_from_buffer_offset(ctx: zengl.Context):
    img = ctx.image((4, 4), 'rgba8unorm', b'AAAA' * 16)
    buf = ctx.buffer(b'UUUUUUUU' + b'BBBB' * 16 + b'VVVVVVVV')
    img.write(buf.view(64, 8))
    assert img.read() == b'BBBB' * 16


def test_image_read_to_buffer(ctx: zengl.Context):
    img = ctx.image((4, 4), 'rgba8unorm', b'AAAA' * 16)
    buf = ctx.buffer(b'BBBB' * 16)
    img.read(into=buf)
    assert buf.read() == b'AAAA' * 16


def test_image_read_to_buffer_offset(ctx: zengl.Context):
    img = ctx.image((4, 4), 'rgba8unorm', b'AAAA' * 16)
    buf = ctx.buffer(b'UUUUUUUU' + b'BBBB' * 16 + b'VVVVVVVV')
    img.read(into=buf.view(64, 8))
    assert buf.read() == b'UUUUUUUU' + b'AAAA' * 16 + b'VVVVVVVV'



import os
import re
import sys

from setuptools import Extension, setup

extra_compile_args = []
extra_link_args = []
define_macros = []

if sys.platform.startswith('darwin'):
    extra_compile_args += ['-Wno-writable-strings']

if os.getenv('PYODIDE'):
    with open('zengl.js') as zengl_js:
        code = re.sub(r'\s+', ' ', zengl_js.read())
        define_macros += [('EXTERN_GL', f'"{code}"')]

if os.getenv('ZENGL_COVERAGE'):
    extra_compile_args += ['-O0', '--coverage']
    extra_link_args += ['-O0', '--coverage']

if os.getenv('ZENGL_WARNINGS'):
    if sys.platform.startswith('linux'):
        extra_compile_args += [
            '-g3',
            '-Wall',
            '-Wextra',
            '-Wconversion',
            '-Wdouble-promotion',
            '-Wno-unused-parameter',
            '-Wno-cast-function-type',
            '-Werror',
        ]
        extra_link_args += [
            '-fsanitize=undefined',
        ]

ext = Extension(
    name='zengl',
    sources=['zengl.c'],
    extra_compile_args=extra_compile_args,
    extra_link_args=extra_link_args,
    define_macros=define_macros,
)

with open('README.md') as readme:
    long_description = readme.read()

setup(
    name='zengl',
    version='2.5.1',
    ext_modules=[ext],
    py_modules=['_zengl'],
    license='MIT',
    platforms=['any'],
    description='OpenGL Pipelines for Python',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Szabolcs Dombi',
    author_email='szabolcs@szabolcsdombi.com',
    url='https://github.com/szabolcsdombi/zengl/',
    project_urls={
        'Documentation': 'https://zengl.readthedocs.io/',
        'Source': 'https://github.com/szabolcsdombi/zengl/',
        'Bug Tracker': 'https://github.com/szabolcsdombi/zengl/issues/',
    },
    keywords=[
        'OpenGL',
        'rendering',
        'graphics',
        'shader',
        'gpu',
    ],
    packages=['zengl-stubs'],
    package_data={'zengl-stubs': ['__init__.pyi']},
    include_package_data=True,
)
